require 'git'
require 'yaml'
require 'rfuse'
require 'gitroom/models/node'
require 'gitroom/models/log'
require 'test/unit/assertions'
require 'gitroom/concerns/node_cached'

module GitRoom
  class Mounter
    include Test::Unit::Assertions
    include NodeCached

    def readdir(ctx, path, filler, offset, ffi)
      find_node path
      @node.nodes.each do |node|
        filler.push node.name, node.stat, 0
      end
    end

    def getattr(ctx, path)
      find_node path
      begin
        @node.stat
      rescue
        print "Unexpected in stat #{path}"
        enoent(path)
      end
    end

    def mkdir(ctx, path, mode)
      update_node path, mode
      cached_node_present! path
    end

    def mknod(ctx, path, mode, major, minor)
      update_node path, mode, ctx.uid, ctx.gid, false, ''
      cached_node_present! path
    end

    def open(ctx, path, ffi)
    end

    def chmod(ctx, path, mode)
      find_node path
      @node.update_attributes! :mode => mode
    end

    def chown(ctx, path, uid, gid)
      find_node path
      @node.update_attributes! :uid => uid, :gid => gid
    end

    def truncate(ctx, path, offset)
      find_node path
      # @node.update_attributes! :content => @node.content[0..offset]
      @node.truncate_content offset
    end

    def utime(ctx, path, actime, modtime)
      find_node path
      @node.update_attributes! :updated_at => Time.at([actime, modtime].max)
    end

    def unlink(ctx, path)
      find_node path
      @node.destroy
      cached_node_absent! path
    end

    def rmdir(ctx, path)
      find_node path
      @node.destroy
      cached_node_absent! path
    end

    # TODO:: Make symlink and link methods

    # def symlink(ctx, path, as)
    # end

    # def link(ctx, path, as)
    # end

    def rename(ctx, path, as)
      find_node path
      move_node(@node, as)
      cached_node_absent! path
      cached_node_absent! as
      cached_node_present! as
    end

    def read(ctx, path, size, offset, fi)
      find_node path
      eisdir(path) if @node.directory
      res = ''
      begin
        res = read_cached_node @node, size, offset
      rescue
        print "Unexpected in read #{path}: #{$!}\n"
        enoent(path)
      end
      res
    end

    def write(ctx, path, buf, offset, fi)
      find_node path
      eisdir(path) if @node.directory
      begin
        write_cached_node @node, buf, offset
      rescue
        print "Unexpected in write #{path}: #{$!}\n"
        enoent(path)
      end
      buf.size
    end

    def setxattr(ctx, path, name, value, size, flags)
      find_node path
      @node.setxattr name, value, flags
    end

    def getxattr(ctx, path, name)
      find_node path
      @node.getxattr(name) || ''
    end

    def listxattr(ctx, path)
      find_node path
      @node.listxattr
    end

    def removexattr(ctx, path, name)
      find_node path
      @node.removexattr name
    end


    def statfs(ctx,path)
      s = RFuse::StatVfs.new
      s.f_bsize    = 1024
      s.f_frsize   = 1024
      s.f_blocks   = 1000000
      s.f_bfree    = 500000
      s.f_bavail   = 990000
      s.f_files    = 10000
      s.f_ffree    = 9900
      s.f_favail   = 9900
      s.f_fsid     = 23423
      s.f_flag     = 0
      s.f_namemax  = 10000
      return s
    end

    def ioctl(ctx, path, cmd, arg, ffi, flags, data)
      # FT: I was not been able to test it.
      print "*** IOCTL: command: ", cmd, "\n"
    end

    def poll(ctx, path, ffi, ph, reventsp)
      print "*** POLL: ", path, "\n"
      # This is how we notify the caller if something happens:
      ph.notifyPoll
      # when the GC harvests the object it calls fuse_pollhandle_destroy
      # by itself.
    end

    def init(ctx,rfuseconninfo)
      print "RFuse TestFS started\n"
      print "init called\n"
      print "proto_major:#{rfuseconninfo.proto_major}\n"
    end

    private

    def find_node(path, ignore = false)
      @node = nil
      # @node = Node.where(:path => path).first
      @node = find_cached_node(path)
      enoent(path) if @node.nil? && !ignore
      @node
    end

    def update_node(path, mode, uid = nil, gid = nil, directory = true, content = nil)
      nodes = split_dirnode(path)
      uid ||= nodes[:dirnode].uid
      gid ||= nodes[:dirnode].gid
      nodes[:node] ||= if directory
        Node.create! :name => nodes[:name], :path => path, :directory => true,
                     :node_id => nodes[:dirnode].id, :mode => mode,
                     :uid => uid, :gid => gid
      else
        Node.create! :name => nodes[:name], :path => path, :directory => false,
                         :node_id => nodes[:dirnode].id, :mode => mode
      end
    end

    def split_dirpath(path)
      pathname = Pathname.new path
      { :name => pathname.basename.to_s, :dirname => pathname.dirname.to_s }
    end

    def split_dirnode(path)
      names = split_dirpath path
      node = find_node path, true
      dirnode = find_node names[:dirname]
      enotdir(path) unless dirnode.directory
      { :node => node, :dirnode => dirnode, :name => names[:name], :dirname => names[:dirname] }
    end

    def move_node(node, path)
      nodes = split_dirnode(path)
      nodes[:node] && nodes[:node].destroy
      node.update_attributes! :path => path, :node_id => nodes[:dirnode].id, :name => nodes[:name]
    end

    def enoent(path = nil)
      # print "Raising enoent #{path}\n"
      raise Errno::ENOENT.new(path)
    end

    def enotdir(path = nil)
      # print "Raising enotdir #{path}\n"
      raise Errno::ENOTDIR.new(path)
    end

    def eisdir(path = nil)
      # print "Raising eisdir #{path}\n"
      raise Errno::EISDIR.new(path)
    end
  end
end
